#
# Copyright 2018 Agilx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Apptuit Pyformance Reporter
"""
import os
import socket
import sys
import time

from pyformance import MetricsRegistry
from pyformance.reporters.reporter import Reporter

from apptuit.apptuit_client import Apptuit, DataPoint, ApptuitSendException, TimeSeriesName
from .process_metrics import ProcessMetrics
from ..utils import _get_tags_from_environment, strtobool

NUMBER_OF_TOTAL_POINTS = "apptuit.reporter.send.total"
NUMBER_OF_SUCCESSFUL_POINTS = "apptuit.reporter.send.successful"
NUMBER_OF_FAILED_POINTS = "apptuit.reporter.send.failed"
API_CALL_TIMER = "apptuit.reporter.send.time"
DISABLE_HOST_TAG = "APPTUIT_DISABLE_HOST_TAG"
BATCH_SIZE = 50000


def default_error_handler(status_code, successful, failed, errors):
    """
    This is the default error handler for the ApptuitReporter.
    It simply writes the errors to stderr.
    Parameters
    ----------
        status_code: response status_code of Apptuit.send()
        successful: number of datapoints updated successfully
        failed: number of datapoints updating failed
        errors: errors in response
    """
    msg = "%d points out of %d had errors\n" \
          "HTTP status returned from Apptuit: %d\n" \
          "Detailed error messages: %s\n" % \
          (failed, successful + failed, status_code, str(errors))
    sys.stderr.write(msg)


class ApptuitReporter(Reporter):
    """
        Pyformance based reporter for Apptuit. It provides high level
        primitives, such as meter, counter, gauge, etc., for collecting
        data and reports them asynchronously to Apptuit.
    """

    def __init__(self, registry=None, reporting_interval=10, token=None,
                 api_endpoint="https://api.apptuit.ai", prefix="", tags=None,
                 error_handler=default_error_handler, disable_host_tag=None,
                 collect_process_metrics=False, sanitize_mode="prometheus",
                 retry_count=0):
        """
        Parameters
        ----------
            registry: An instance of MetricsRegistry from pyformance. It is
                used as a container for all the metrics. If None, a new instance will be
                created internally
            reporting_interval: Reporting interval in seconds
            token: Apptuit API token
            prefix: Optional prefix for metric names, this will be prepended to all the
                metric names
            tags: A dictionary of tag keys and values which will be included with
                all the metrics reported by this reporter
            error_handler: A function to be executed in case of errors when reporting
                the data. If not specified, the default error handler will be used which by
                default writes the errors to stderr. The expected signature of an error handler
                is: error_handler(status_code, successful_points, failed_points, errors). Here
                status_code is the HTTP status code of the failed API call, successful_points is
                number of points processed successfully, failed_points is number of failed points
                and errors is a list of error messages describing reason of each failure.
            disable_host_tag: By default a host tag will be added to all the metrics reported by
                the reporter. Set disable_host_tag to False if you wish to disable it
            collect_process_metrics: A boolean variable specifying if process metrics should be
                collected or not, if set to True then process metrics will be collected. By default,
                this will collect resource, thread, and gc metrics.
            sanitize_mode: Is a string value which will enable sanitizer, sanitizer will
                    automatically change your metric names to be compatible with apptuit
                    or prometheus. Set it to None if not needed.
            retry_count: This will allow you to retry to send DP's in case of errors.
                This uses Backoff-jitter algo to retry.
                `https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/`
        """
        super(ApptuitReporter, self).__init__(registry=registry,
                                              reporting_interval=reporting_interval)
        self.retry_count = retry_count
        self.endpoint = api_endpoint
        self.token = token
        self.tags = tags
        environ_tags = _get_tags_from_environment()
        if environ_tags:
            if self.tags is not None:
                environ_tags.update(self.tags)
            self.tags = environ_tags
        if disable_host_tag is None:
            disable_host_tag = os.environ.get(DISABLE_HOST_TAG, False)
            if disable_host_tag:
                disable_host_tag = strtobool(disable_host_tag)

        if not disable_host_tag:
            if self.tags:
                if self.tags.get("host") is None:
                    self.tags["host"] = socket.gethostname()
            else:
                self.tags = {"host": socket.gethostname()}
        self.prefix = prefix if prefix is not None else ""
        self.client = Apptuit(token=token, api_endpoint=api_endpoint,
                              ignore_environ_tags=True, sanitize_mode=sanitize_mode)
        self._meta_metrics_registry = MetricsRegistry()
        self.error_handler = error_handler
        self.process_metrics = None
        if collect_process_metrics:
            self.process_metrics = ProcessMetrics(self.registry)

    def _update_counter(self, key, value):
        """
        To increment the counter with `key` by `value`.
        :param key: Name of counter.
        :param value: value to increment.
        """
        self._meta_metrics_registry.counter(key).inc(value)

    def report_now(self, registry=None, timestamp=None):
        """
        Report the data
        Params:
            registry: pyformance Registry containing all metrics
            timestamp: timestamp of the data point
        """
        if self.process_metrics:
            self.process_metrics.collect_process_metrics()
        dps = self._collect_data_points(registry or self.registry, timestamp)
        meta_dps = self._collect_data_points(self._meta_metrics_registry)
        if not dps:
            return
        dps_len = len(dps)
        success_count = 0
        failed_count = 0
        errors = []
        for i in range(0, dps_len, BATCH_SIZE):
            try:
                with self._meta_metrics_registry.timer(API_CALL_TIMER).time():
                    end_index = min(dps_len, i + BATCH_SIZE)
                    self.client.send(dps[i: end_index], self.retry_count)
                    points_sent_count = end_index - i
                    self._update_counter(NUMBER_OF_TOTAL_POINTS, points_sent_count)
                    self._update_counter(NUMBER_OF_SUCCESSFUL_POINTS, points_sent_count)
                    self._update_counter(NUMBER_OF_FAILED_POINTS, 0)
                    success_count += points_sent_count
            except ApptuitSendException as exception:
                self._update_counter(NUMBER_OF_SUCCESSFUL_POINTS, exception.success)
                self._update_counter(NUMBER_OF_FAILED_POINTS, exception.failed)
                success_count += exception.success
                failed_count += exception.failed
                errors += exception.errors
                if self.error_handler:
                    self.error_handler(
                        exception.status_code,
                        exception.success,
                        exception.failed,
                        exception.errors
                    )
        self.client.send(meta_dps, self.retry_count)
        if failed_count != 0:
            raise ApptuitSendException("Failed to send %d out of %d points" %
                                       (failed_count, dps_len), success=success_count,
                                       failed=failed_count, errors=errors)

    @staticmethod
    def _get_tags(key):
        """
        Get tags of a metric
        Params:
            metric key
        Returns:
            metric name, dictionary of tags
        """
        metric_name, metric_tags = TimeSeriesName.decode_metric(key)
        return metric_name, metric_tags

    def _collect_data_points(self, registry, timestamp=None):
        """
        will collect all metrics from registry and convert them to DataPoints
        Params:
            registry: pyformance registry object
            timestamp: timestamp of the data point
        Returns:
            list of DataPoints
        """
        timestamp = timestamp or int(round(self.clock.time()))
        metrics = registry.dump_metrics()
        dps = []
        global_tags = self.tags if self.tags else {}
        for key in metrics.keys():
            metric_name, metric_tags = self._get_tags(key)
            if metric_tags and global_tags:
                tags = global_tags.copy()
                tags.update(metric_tags)
            elif metric_tags:
                tags = metric_tags
            else:
                tags = global_tags
            for value_key in metrics[key].keys():
                data_point = DataPoint(
                    metric=self.prefix + metric_name + '.' + value_key,
                    tags=tags, timestamp=timestamp, value=metrics[key][value_key])
                dps.append(data_point)
        return dps

    def _loop(self):
        while not self._stopped.is_set():
            start = time.time()
            try:
                self.report_now(self.registry)
            except Exception:
                pass
            end = time.time()
            wait_time = max(0, self.reporting_interval-(end-start))
            time.sleep(wait_time)



#
# Copyright 2018 Agilx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Peocess Metrics Collection
"""
import gc
import os
import resource
import threading

from apptuit.apptuit_client import TimeSeriesName

RESOURCE_STRUCT_RUSAGE = ["ru_utime", "ru_stime",
                          "ru_maxrss", "ru_ixrss",
                          "ru_idrss", "ru_isrss",
                          "ru_minflt", "ru_majflt",
                          "ru_nswap", "ru_inblock",
                          "ru_oublock", "ru_msgsnd",
                          "ru_msgrcv", "ru_nsignals",
                          "ru_nvcsw", "ru_nivcsw"
                          ]


class ProcessMetrics(object):

    def __init__(self, registry):
        self.pid = os.getpid()
        self.registry = registry
        self.resource_metric_names = self._get_resource_metic_names()
        self.thread_metrics_names = self._get_thread_metic_names()
        self.gc_metric_names = self._get_gc_metric_names()
        self.previous_resource_metrics = [0] * len(self.resource_metric_names)
        self.previous_gc_metrics = [0] * len(self.gc_metric_names)

    def collect_process_metrics(self):
        """
        To collect all the process metrics.
        """
        rm = resource.getrusage(resource.RUSAGE_SELF)
        resource_metrics = {}
        for res_name in RESOURCE_STRUCT_RUSAGE:
            resource_metrics[res_name] = getattr(rm, res_name)
        for res_name in RESOURCE_STRUCT_RUSAGE[2:6]:
            resource_metrics[res_name] = resource_metrics[res_name] * 1024
        resource_metrics = [cur_val - pre_val
                            for cur_val, pre_val in
                            zip([resource_metrics[key]
                                 for key in resource_metrics], self.previous_resource_metrics)]
        self._collect_counter_from_list(self.resource_metric_names, resource_metrics)
        th_values = threading.enumerate()
        thread_metrics = [
            [t.daemon is True for t in th_values].count(True),
            [t.daemon is False for t in th_values].count(True),
            [isinstance(t, threading._DummyThread) for t in th_values].count(True)
        ]
        self._collect_gauge_from_list(self.thread_metrics_names, thread_metrics)
        if gc.isenabled():
            collection = list(gc.get_count())
            threshold = list(gc.get_threshold())
            gc_metrics = collection + threshold
            gc_metrics = [cur_val - pre_val
                          for cur_val, pre_val in zip(gc_metrics, self.previous_gc_metrics)]
            self._collect_counter_from_list(self.gc_metric_names, gc_metrics)
            self.previous_gc_metrics = gc_metrics
        self.previous_resource_metrics = resource_metrics

    def _get_gc_metric_names(self):
        """
        To get a list of gc metric names.
        :return: a list of gc metric names.
        """
        gc_metric_names = [
            TimeSeriesName.encode_metric("python.gc.collection",
                                         {"type": "collection_0", "worker_id": self.pid}),
            TimeSeriesName.encode_metric("python.gc.collection",
                                         {"type": "collection_1", "worker_id": self.pid}),
            TimeSeriesName.encode_metric("python.gc.collection",
                                         {"type": "collection_2", "worker_id": self.pid}),
            TimeSeriesName.encode_metric("python.gc.threshold",
                                         {"type": "threshold_0", "worker_id": self.pid}),
            TimeSeriesName.encode_metric("python.gc.threshold",
                                         {"type": "threshold_1", "worker_id": self.pid}),
            TimeSeriesName.encode_metric("python.gc.threshold",
                                         {"type": "threshold_2", "worker_id": self.pid})
        ]
        return gc_metric_names

    def _get_thread_metic_names(self):
        """
        To get a list of thread metric names.
        :return: a list of thread metric names.
        """
        thread_metric_names = [
            TimeSeriesName.encode_metric("python.thread",
                                         {"type": "demon", "worker_id": self.pid}),
            TimeSeriesName.encode_metric("python.thread",
                                         {"type": "alive", "worker_id": self.pid}),
            TimeSeriesName.encode_metric("python.thread",
                                         {"type": "dummy", "worker_id": self.pid}),
        ]
        return thread_metric_names

    def _get_resource_metic_names(self):
        """
        To get a list of resource metric names.
        :return: a list of resource metric names.
        """
        resource_metric_names = [
            TimeSeriesName.encode_metric("python.cpu.time.used.seconds",
                                         {"type": "user", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.cpu.time.used.seconds",
                                         {"type": "system", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.memory.usage.bytes",
                                         {"type": "main", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.memory.usage.bytes",
                                         {"type": "shared", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.memory.usage.bytes",
                                         {"type": "unshared", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.memory.usage.bytes",
                                         {"type": "unshared_stack_size",
                                          "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.page.faults",
                                         {"type": "major", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.page.faults",
                                         {"type": "minor", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.process.swaps",
                                         {"worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.block.operations",
                                         {"type": "input", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.block.operations",
                                         {"type": "output", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.ipc.messages",
                                         {"type": "sent", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.ipc.messages",
                                         {"type": "received", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.system.signals",
                                         {"worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.context.switches",
                                         {"type": "voluntary", "worker_id": self.pid, }),
            TimeSeriesName.encode_metric("python.context.switches",
                                         {"type": "involuntary", "worker_id": self.pid, }),
        ]
        return resource_metric_names

    def _collect_counter_from_list(self, metric_names, metric_values):
        """
        To increment list of counters `metric_names` with values `metric_values`.
        :param metric_names: A list of counter names.
        :param metric_values: A list of corresponding value to increment.
        """
        for ind, metric_value in enumerate(metric_values):
            metric_counter = self.registry.counter(metric_names[ind])
            metric_counter.inc(metric_value)

    def _collect_gauge_from_list(self, metric_names, metric_values):
        """
        To increment list of gauge `metric_names` with values `metric_values`.
        :param metric_names: A list of gauge names.
        :param metric_values: A list of corresponding value to set.
        """
        for ind, metric in enumerate(metric_values):
            metric_counter = self.registry.gauge(metric_names[ind])
            metric_counter.set_value(metric)



#
# Copyright 2018 Agilx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
utilises for apptuit
"""
import os
import re
import warnings
from string import ascii_letters, digits

from apptuit import APPTUIT_PY_TAGS, DEPRECATED_APPTUIT_PY_TAGS

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache

VALID_REGEX = re.compile(r"[-\w_/.]+$", re.U)
PROMETHEUS_VALID_CHARSET = set(ascii_letters + digits + "_")
APPTUIT_SANITIZE_REGEX = re.compile(r'([^-\w_./])', re.U)
REPLACE_WITH_SINGLE_UNDERSCORE_REGEX = re.compile('_+')


@lru_cache(maxsize=2048)
def sanitize_name_prometheus(name):
    """
    To make the metric name Prometheus compatible.
    :param name: a string value metric name or tag-key.
    :return: metric_name which is Prometheus compatible.
    """
    if name[0] in digits:
        name = "_" + name
    invalid_char_set = set(name).difference(PROMETHEUS_VALID_CHARSET)
    for invalid_char in invalid_char_set:
        if invalid_char in name:
            name = name.replace(invalid_char, "_")
    # to replace multiple '_' to one
    return REPLACE_WITH_SINGLE_UNDERSCORE_REGEX.sub("_", name)


@lru_cache(maxsize=2048)
def sanitize_name_apptuit(name):
    """
    To make the metric name Apptuit compatible.
    :param name: a string value metric name or tag-key.
    :return: metric_name which is Apptuit compatible.
    """
    substitute_string = APPTUIT_SANITIZE_REGEX.sub("_", name)
    # to replace multiple '_' to one
    return REPLACE_WITH_SINGLE_UNDERSCORE_REGEX.sub("_", substitute_string)


@lru_cache(maxsize=4096)
def _contains_valid_chars(string):
    return VALID_REGEX.match(string) is not None


def _validate_tags(tags):
    for tagk in tags.keys():
        if not tagk or not _contains_valid_chars(tagk):
            raise ValueError("Tag key %s contains an invalid character, "
                             "allowed characters are a-z, A-Z, 0-9, -, _, ., and /" % tagk)


def _get_tags_from_environment():
    tags_str = os.environ.get(APPTUIT_PY_TAGS)
    if not tags_str:
        tags_str = os.environ.get(DEPRECATED_APPTUIT_PY_TAGS)
        if tags_str:
            warnings.warn("The environment variable %s is deprecated, please use %s instead"
                          % (DEPRECATED_APPTUIT_PY_TAGS, APPTUIT_PY_TAGS), DeprecationWarning)
    if not tags_str:
        return {}
    tags = {}
    tags_str = tags_str.strip(", ")
    tags_split = tags_str.split(',')
    for tag in tags_split:
        tag = tag.strip()
        if not tag:
            continue
        try:
            key, val = tag.split(":")
            tags[key.strip()] = val.strip()
        except ValueError:
            raise ValueError("Invalid format for " + APPTUIT_PY_TAGS +
                             ", failed to parse tag key-value pair '" + tag + "', " +
                             APPTUIT_PY_TAGS + " should be in the format - "
                                               "'tag_key1:tag_val1,tag_key2:tag_val2"
                                               ",...,tag_keyN:tag_valN'")
    _validate_tags(tags)
    return tags


def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return 1
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return 0
    else:
        raise ValueError("invalid truth value %r" % (val,))



#
# Copyright 2018 Agilx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

APPTUIT_PY_TOKEN = "APPTUIT_API_TOKEN"
APPTUIT_PY_TAGS = "APPTUIT_TAGS"
DEPRECATED_APPTUIT_PY_TOKEN = "APPTUIT_PY_TOKEN"
DEPRECATED_APPTUIT_PY_TAGS = "APPTUIT_PY_TAGS"
__version__ = '2.4.1'

from apptuit import pyformance, timeseries
from .apptuit_client import Apptuit, DataPoint, ApptuitException, ApptuitSendException, \
    TimeSeriesName, TimeSeries

__all__ = ['Apptuit', 'DataPoint', 'ApptuitException', 'TimeSeriesName', 'TimeSeries',
           'pyformance', 'timeseries', 'ApptuitSendException', '__version__']
